/************************************************************************
**
**  Copyright (C) 2015-2021 Kevin B. Hendricks, Stratford Ontario Canada
**  Copyright (C) 2014      John Schember <john@nachtimwald.com> 
**
**  This file is part of Sigil.
**
**  Sigil is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  Sigil is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with Sigil.  If not, see <http://www.gnu.org/licenses/>.
**
*************************************************************************/

#include <QString>

#include "Misc/ValidationResult.h"

ValidationResult::ValidationResult(ValidationResult::ResType type, const QString &bookpath, int linenumber, const QString &message)
{
    m_type       = type;
    m_bookpath   = bookpath;
    m_linenumber = linenumber;
    m_charoffset = -1;
    m_message    = message;
}

ValidationResult::ValidationResult(ValidationResult::ResType type, const QString &bookpath, int linenumber, int charoffset, const QString &message)
{
    m_type       = type;
    m_bookpath   = bookpath;
    m_linenumber = linenumber;
    m_charoffset = charoffset;
    m_message    = message;
}

ValidationResult::~ValidationResult()
{
}

ValidationResult::ResType ValidationResult::Type()
{
    return m_type;
}

QString ValidationResult::BookPath()
{
    return m_bookpath;
}

int ValidationResult::LineNumber()
{
    return m_linenumber;
}

int ValidationResult::CharOffset()
{
    return m_charoffset;
}

QString ValidationResult::Message()
{
    return m_message;
}
