/************************************************************************
**
**  Copyright (C) 2020-2024 Doug Massay
**
**  This file is part of Sigil.
**
**  Sigil is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  Sigil is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with Sigil.  If not, see <http://www.gnu.org/licenses/>.
**
*************************************************************************/

#pragma once
#ifndef SIGILDARKSTYLE_H
#define SIGILDARKSTYLE_H

#include <QApplication>
#include <QFile>
#include <QProxyStyle>
#include <QStyleFactory>

class SigilDarkStyle : public QProxyStyle {
    Q_OBJECT

public:
    SigilDarkStyle();
    explicit SigilDarkStyle(QStyle *style);

    QStyle *baseStyle() const;

    void polish(QPalette &palette) override;
    void polish(QApplication *app) override;
    void unpolish(QApplication *app) override;

private:
    QStyle *styleBase(QStyle *style = Q_NULLPTR) const;
};

#endif  // SIGILDARKSTYLE_H

