// Do not edit
// Generated from tag.in (see gentags.py)

"a",
"abbr",
"acronym",
"address",
"altglyph",
"altglyphdef",
"altglyphitem",
"animate",
"animatecolor",
"animatemotion",
"animatetransform",
"annotation-xml",
"applet",
"area",
"article",
"aside",
"audio",
"b",
"base",
"basefont",
"bdi",
"bdo",
"bgsound",
"big",
"blink",
"blockquote",
"body",
"br",
"button",
"canvas",
"caption",
"center",
"circle",
"cite",
"clippath",
"code",
"col",
"colgroup",
"color-profile",
"cursor",
"data",
"datalist",
"dd",
"defs",
"del",
"desc",
"details",
"dfn",
"dialog",
"dir",
"div",
"dl",
"dt",
"ellipse",
"em",
"embed",
"feblend",
"fecolormatrix",
"fecomponenttransfer",
"fecomposite",
"feconvolvematrix",
"fediffuselighting",
"fedisplacementmap",
"fedistantlight",
"feflood",
"fefunca",
"fefuncb",
"fefuncg",
"fefuncr",
"fegaussianblur",
"feimage",
"femerge",
"femergenode",
"femorphology",
"feoffset",
"fepointlight",
"fespectactualrlighting",
"fespotlight",
"fetile",
"feturbulence",
"fieldset",
"figcaption",
"figure",
"filter",
"font",
"font-face",
"font-face-format",
"font-face-name",
"font-face-src",
"font-face-uri",
"footer",
"foreignobject",
"form",
"frame",
"frameset",
"g",
"glyph",
"glyphref",
"h1",
"h2",
"h3",
"h4",
"h5",
"h6",
"head",
"header",
"hgroup",
"hkern",
"hr",
"html",
"i",
"iframe",
"image",
"img",
"input",
"ins",
"isindex",
"kbd",
"keygen",
"label",
"legend",
"li",
"line",
"lineargradient",
"link",
"listing",
"maction",
"main",
"maligngroup",
"malignmark",
"map",
"mark",
"marker",
"marquee",
"mask",
"math",
"menclose",
"menu",
"menuitem",
"merror",
"meta",
"metadata",
"meter",
"mfenced",
"mfrac",
"mglyph",
"mi",
"missing-glyph",
"mlabeledtr",
"mlongdiv",
"mmultiscripts",
"mn",
"mo",
"mover",
"mpadded",
"mpath",
"mphantom",
"mprescripts",
"mroot",
"mrow",
"ms",
"mscarries",
"mscarry",
"msgroup",
"msline",
"mspace",
"msqrt",
"msrow",
"mstack",
"mstyle",
"msub",
"msubsup",
"msup",
"mtable",
"mtd",
"mtext",
"mtr",
"multicol",
"munder",
"munderover",
"nav",
"nextid",
"nobr",
"noembed",
"noframes",
"none",
"noscript",
"object",
"ol",
"optgroup",
"option",
"output",
"p",
"param",
"path",
"pattern",
"picture",
"plaintext",
"polygon",
"polyline",
"pre",
"progress",
"q",
"radialgradient",
"rb",
"rect",
"rp",
"rt",
"rtc",
"ruby",
"s",
"samp",
"script",
"search",
"section",
"select",
"semantics",
"set",
"slot",
"small",
"source",
"spacer",
"span",
"stop",
"strike",
"strong",
"style",
"sub",
"summary",
"sup",
"svg",
"switch",
"symbol",
"table",
"tbody",
"td",
"template",
"text",
"textpath",
"textarea",
"tfoot",
"th",
"thead",
"time",
"title",
"tr",
"track",
"tref",
"tspan",
"tt",
"u",
"ul",
"use",
"var",
"video",
"view",
"vkern",
"wbr",
"xmp",
